@@ -25,24 +25,24 @@
 #endif
 
 #include <string.h>
+#include <alloca.h>
 #include <linux/if_ether.h>
 #include <ell/ell.h>
 
-#include "crypto.h"
-#include "eapol.h"
-#include "ie.h"
-#include "util.h"
-#include "mpdu.h"
-#include "eap.h"
-#include "handshake.h"
-#include "watchlist.h"
+#include "src/crypto.h"
+#include "src/eapol.h"
+#include "src/ie.h"
+#include "src/util.h"
+#include "src/mpdu.h"
+#include "src/eap.h"
+#include "src/handshake.h"
+#include "src/watchlist.h"
 
 struct l_queue *state_machines;
 struct l_queue *preauths;
 struct watchlist frame_watches;
 static uint32_t eapol_4way_handshake_time = 2;
 
-eapol_deauthenticate_func_t deauthenticate = NULL;
 eapol_rekey_offload_func_t rekey_offload = NULL;
 
 eapol_tx_packet_func_t tx_packet = NULL;
@@ -65,8 +65,8 @@ uint32_t next_frame_watch_id;
  *
  * The input struct eapol_key *frame should have a zero-d MIC field
  */
-bool eapol_calculate_mic(const uint8_t *kck, const struct eapol_key *frame,
-				uint8_t *mic)
+bool eapol_calculate_mic(enum ie_rsn_akm_suite akm, const uint8_t *kck,
+				const struct eapol_key *frame, uint8_t *mic)
 {
 	size_t frame_len = sizeof(struct eapol_key);
 
@@ -79,12 +79,21 @@ bool eapol_calculate_mic(const uint8_t *kck, const struct eapol_key *frame,
 		return hmac_sha1(kck, 16, frame, frame_len, mic, 16);
 	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
 		return cmac_aes(kck, 16, frame, frame_len, mic, 16);
+	case EAPOL_KEY_DESCRIPTOR_VERSION_AKM_DEFINED:
+		switch (akm) {
+		case IE_RSN_AKM_SUITE_SAE_SHA256:
+		case IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256:
+			return cmac_aes(kck, 16, frame, frame_len, mic, 16);
+		default:
+			return false;
+		}
 	default:
 		return false;
 	}
 }
 
-bool eapol_verify_mic(const uint8_t *kck, const struct eapol_key *frame)
+bool eapol_verify_mic(enum ie_rsn_akm_suite akm, const uint8_t *kck,
+			const struct eapol_key *frame)
 {
 	size_t frame_len = sizeof(struct eapol_key);
 	uint8_t mic[16];
@@ -112,6 +121,17 @@ bool eapol_verify_mic(const uint8_t *kck, const struct eapol_key *frame)
 		break;
 	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
 		checksum = l_checksum_new_cmac_aes(kck, 16);
+		break;
+	case EAPOL_KEY_DESCRIPTOR_VERSION_AKM_DEFINED:
+		switch (akm) {
+		case IE_RSN_AKM_SUITE_SAE_SHA256:
+		case IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256:
+			checksum = l_checksum_new_cmac_aes(kck, 16);
+			break;
+		default:
+			return false;
+		}
+
 		break;
 	default:
 		return false;
@@ -130,7 +150,7 @@ bool eapol_verify_mic(const uint8_t *kck, const struct eapol_key *frame)
 	return false;
 }
 
-uint8_t *eapol_decrypt_key_data(const uint8_t *kek,
+uint8_t *eapol_decrypt_key_data(enum ie_rsn_akm_suite akm, const uint8_t *kek,
 				const struct eapol_key *frame,
 				size_t *decrypted_size)
 {
@@ -143,8 +163,22 @@ uint8_t *eapol_decrypt_key_data(const uint8_t *kek,
 	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_MD5_ARC4:
 		expected_len = key_data_len;
 		break;
+	case EAPOL_KEY_DESCRIPTOR_VERSION_AKM_DEFINED:
+		/*
+		 * TODO: for now, only SAE is supported under the AKM_DEFINED
+		 * key descriptor version. Once 8021x suites are added for this
+		 * type this will need to be expanded to handle the AKM types in
+		 * its own switch.
+		 */
+		if (!IE_AKM_IS_SAE(akm))
+			return NULL;
+
+		/* Fall through */
 	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES:
 	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
+		if (key_data_len < 24 || key_data_len % 8)
+			return NULL;
+
 		expected_len = key_data_len - 8;
 		break;
 	default:
@@ -172,9 +206,7 @@ uint8_t *eapol_decrypt_key_data(const uint8_t *kek,
 	}
 	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES:
 	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
-		if (key_data_len < 24 || key_data_len % 8)
-			goto error;
-
+	case EAPOL_KEY_DESCRIPTOR_VERSION_AKM_DEFINED:
 		if (!aes_unwrap(kek, key_data, key_data_len, buf))
 			goto error;
 
@@ -199,7 +231,7 @@ error:
  * Note that for efficiency @key_data is being modified, including in
  * case of failure, so it must be sufficiently larger than @key_data_len.
  */
-bool eapol_encrypt_key_data(const uint8_t *kek, uint8_t *key_data,
+static bool eapol_encrypt_key_data(const uint8_t *kek, uint8_t *key_data,
 				size_t key_data_len,
 				struct eapol_key *out_frame)
 {
@@ -228,7 +260,8 @@ bool eapol_encrypt_key_data(const uint8_t *kek, uint8_t *key_data,
 	return true;
 }
 
-void eapol_key_data_append(struct eapol_key *ek, enum handshake_kde selector,
+static void eapol_key_data_append(struct eapol_key *ek,
+				enum handshake_kde selector,
 				const uint8_t *data, size_t data_len)
 {
 	uint16_t key_data_len = L_BE16_TO_CPU(ek->key_data_len);
@@ -244,52 +277,6 @@ void eapol_key_data_append(struct eapol_key *ek, enum handshake_kde selector,
 	ek->key_data_len = L_CPU_TO_BE16(key_data_len);
 }
 
-const struct eapol_key *eapol_key_validate(const uint8_t *frame, size_t len)
-{
-	const struct eapol_key *ek;
-	uint16_t key_data_len;
-
-	if (len < sizeof(struct eapol_key))
-		return NULL;
-
-	ek = (const struct eapol_key *) frame;
-
-	switch (ek->header.protocol_version) {
-	case EAPOL_PROTOCOL_VERSION_2001:
-	case EAPOL_PROTOCOL_VERSION_2004:
-		break;
-	default:
-		return NULL;
-	}
-
-	if (ek->header.packet_type != 3)
-		return NULL;
-
-	switch (ek->descriptor_type) {
-	case EAPOL_DESCRIPTOR_TYPE_RC4:
-	case EAPOL_DESCRIPTOR_TYPE_80211:
-	case EAPOL_DESCRIPTOR_TYPE_WPA:
-		break;
-	default:
-		return NULL;
-	}
-
-	switch (ek->key_descriptor_version) {
-	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_MD5_ARC4:
-	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES:
-	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
-		break;
-	default:
-		return NULL;
-	}
-
-	key_data_len = L_BE16_TO_CPU(ek->key_data_len);
-	if (len < sizeof(struct eapol_key) + key_data_len)
-		return NULL;
-
-	return ek;
-}
-
 #define VERIFY_PTK_COMMON(ek)	\
 	if (!ek->key_type)	\
 		return false;	\
@@ -323,7 +310,6 @@ bool eapol_verify_ptk_1_of_4(const struct eapol_key *ek)
 	if (ek->wpa_key_id)
 		return false;
 
-	VERIFY_IS_ZERO(ek->eapol_key_iv);
 	VERIFY_IS_ZERO(ek->key_rsc);
 	VERIFY_IS_ZERO(ek->reserved);
 	VERIFY_IS_ZERO(ek->key_mic_data);
@@ -356,6 +342,9 @@ bool eapol_verify_ptk_2_of_4(const struct eapol_key *ek)
 	if (ek->wpa_key_id)
 		return false;
 
+	if (ek->request)
+		return false;
+
 	key_len = L_BE16_TO_CPU(ek->key_length);
 	if (key_len != 0)
 		return false;
@@ -405,12 +394,6 @@ bool eapol_verify_ptk_3_of_4(const struct eapol_key *ek, bool is_wpa)
 
 	VERIFY_IS_ZERO(ek->reserved);
 
-	/* 0 (Version 2) or random (Version 1) */
-	if (ek->key_descriptor_version ==
-			EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES)
-		L_WARN_ON(!util_mem_is_zero(ek->eapol_key_iv,
-						sizeof(ek->eapol_key_iv)));
-
 	return true;
 }
 
@@ -439,6 +422,9 @@ bool eapol_verify_ptk_4_of_4(const struct eapol_key *ek, bool is_wpa)
 	if (ek->wpa_key_id)
 		return false;
 
+	if (ek->request)
+		return false;
+
 	key_len = L_BE16_TO_CPU(ek->key_length);
 	if (key_len != 0)
 		return false;
@@ -495,11 +481,6 @@ bool eapol_verify_gtk_1_of_2(const struct eapol_key *ek, bool is_wpa)
 
 	VERIFY_IS_ZERO(ek->reserved);
 
-	/* 0 (Version 2) or random (Version 1) */
-	if (ek->key_descriptor_version ==
-			EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES)
-		VERIFY_IS_ZERO(ek->eapol_key_iv);
-
 	/*
 	 * WPA_80211_v3_1, Section 2.2.4:
 	 * "Key Index (bits 4 and 5): specifies the key id of the temporal
@@ -648,6 +629,41 @@ struct eapol_key *eapol_create_gtk_2_of_2(
 	return step2;
 }
 
+struct eapol_frame_watch {
+	uint32_t ifindex;
+	struct watchlist_item super;
+};
+
+static void eapol_frame_watch_free(struct watchlist_item *item)
+{
+	struct eapol_frame_watch *efw =
+		container_of(item, struct eapol_frame_watch, super);
+
+	l_free(efw);
+}
+
+static const struct watchlist_ops eapol_frame_watch_ops = {
+	.item_free = eapol_frame_watch_free,
+};
+
+static int32_t eapol_frame_watch_add(uint32_t ifindex,
+					eapol_frame_watch_func_t handler,
+					void *user_data)
+{
+	struct eapol_frame_watch *efw;
+
+	efw = l_new(struct eapol_frame_watch, 1);
+	efw->ifindex = ifindex;
+
+	return watchlist_link(&frame_watches, &efw->super,
+				handler, user_data, NULL);
+}
+
+static bool eapol_frame_watch_remove(uint32_t id)
+{
+	return watchlist_remove(&frame_watches, id);
+}
+
 struct eapol_sm {
 	struct handshake_state *handshake;
 	enum eapol_protocol_version protocol_version;
@@ -656,6 +672,8 @@ struct eapol_sm {
 	void *user_data;
 	struct l_timeout *timeout;
 	struct l_timeout *eapol_start_timeout;
+	unsigned int frame_retry;
+	uint16_t listen_interval;
 	bool have_replay:1;
 	bool started:1;
 	bool use_eapol_start:1;
@@ -721,6 +739,11 @@ void eapol_sm_set_protocol_version(struct eapol_sm *sm,
 	sm->protocol_version = protocol_version;
 }
 
+void eapol_sm_set_listen_interval(struct eapol_sm *sm, uint16_t interval)
+{
+	sm->listen_interval = interval;
+}
+
 void eapol_sm_set_user_data(struct eapol_sm *sm, void *user_data)
 {
 	sm->user_data = user_data;
@@ -740,10 +763,7 @@ static void eapol_sm_write(struct eapol_sm *sm, const struct eapol_frame *ef,
 
 static inline void handshake_failed(struct eapol_sm *sm, uint16_t reason_code)
 {
-	if (deauthenticate)
-		deauthenticate(sm->handshake->ifindex,
-				sm->handshake->aa, sm->handshake->spa,
-				reason_code, sm->user_data);
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_FAILED, &reason_code);
 
 	eapol_sm_free(sm);
 }
@@ -799,6 +819,8 @@ static void send_eapol_start(struct l_timeout *timeout, void *user_data)
 	uint8_t buf[sizeof(struct eapol_frame)];
 	struct eapol_frame *frame = (struct eapol_frame *) buf;
 
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_STARTED, NULL);
+
 	l_timeout_remove(sm->eapol_start_timeout);
 	sm->eapol_start_timeout = NULL;
 
@@ -812,6 +834,97 @@ static void send_eapol_start(struct l_timeout *timeout, void *user_data)
 	eapol_sm_write(sm, frame, false);
 }
 
+static void eapol_set_key_timeout(struct eapol_sm *sm,
+					l_timeout_notify_cb_t cb)
+{
+	/*
+	 * 802.11-2016 12.7.6.6: "The retransmit timeout value shall be
+	 * 100 ms for the first timeout, half the listen interval for the
+	 * second timeout, and the listen interval for subsequent timeouts.
+	 * If there is no listen interval or the listen interval is zero,
+	 * then 100 ms shall be used for all timeout values."
+	 */
+	unsigned int timeout_ms = 100;
+	unsigned int beacon_us = 100 * 1024;
+
+	sm->frame_retry++;
+
+	if (sm->frame_retry == 2 &&
+			sm->listen_interval != 0)
+		timeout_ms = sm->listen_interval * beacon_us / 2000;
+	else if (sm->frame_retry > 2 &&
+			sm->listen_interval != 0)
+		timeout_ms = sm->listen_interval * beacon_us / 1000;
+
+	if (sm->frame_retry > 1)
+		l_timeout_modify_ms(sm->timeout, timeout_ms);
+	else {
+		if (sm->timeout)
+			l_timeout_remove(sm->timeout);
+
+		sm->timeout = l_timeout_create_ms(timeout_ms, cb, sm,
+								NULL);
+	}
+}
+
+/* 802.11-2016 Section 12.7.6.2 */
+static void eapol_send_ptk_1_of_4(struct eapol_sm *sm)
+{
+	uint32_t ifindex = sm->handshake->ifindex;
+	const uint8_t *aa = sm->handshake->aa;
+	uint8_t frame_buf[512];
+	struct eapol_key *ek = (struct eapol_key *) frame_buf;
+	enum crypto_cipher cipher = ie_rsn_cipher_suite_to_cipher(
+				sm->handshake->pairwise_cipher);
+	uint8_t pmkid[16];
+
+	handshake_state_new_anonce(sm->handshake);
+
+	sm->handshake->ptk_complete = false;
+
+	sm->replay_counter++;
+
+	memset(ek, 0, sizeof(struct eapol_key));
+	ek->header.protocol_version = sm->protocol_version;
+	ek->header.packet_type = 0x3;
+	ek->descriptor_type = EAPOL_DESCRIPTOR_TYPE_80211;
+	/* Must be HMAC-SHA1-128 + AES when using CCMP with PSK or 8021X */
+	ek->key_descriptor_version = EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES;
+	ek->key_type = true;
+	ek->key_ack = true;
+	ek->key_length = L_CPU_TO_BE16(crypto_cipher_key_len(cipher));
+	ek->key_replay_counter = L_CPU_TO_BE64(sm->replay_counter);
+	memcpy(ek->key_nonce, sm->handshake->anonce, sizeof(ek->key_nonce));
+
+	/* Write the PMKID KDE into Key Data field unencrypted */
+	crypto_derive_pmkid(sm->handshake->pmk, sm->handshake->spa, aa,
+			pmkid, false);
+	eapol_key_data_append(ek, HANDSHAKE_KDE_PMKID, pmkid, 16);
+
+	ek->header.packet_len = L_CPU_TO_BE16(sizeof(struct eapol_key) +
+					L_BE16_TO_CPU(ek->key_data_len) - 4);
+
+	l_debug("STA: "MAC" retries=%u", MAC_STR(sm->handshake->spa),
+			sm->frame_retry);
+
+	__eapol_tx_packet(ifindex, sm->handshake->spa, ETH_P_PAE,
+				(struct eapol_frame *) ek, false);
+}
+
+static void eapol_ptk_1_of_4_retry(struct l_timeout *timeout, void *user_data)
+{
+	struct eapol_sm *sm = user_data;
+
+	if (sm->frame_retry >= 3) {
+		handshake_failed(sm, MMPDU_REASON_CODE_4WAY_HANDSHAKE_TIMEOUT);
+		return;
+	}
+
+	eapol_send_ptk_1_of_4(sm);
+
+	eapol_set_key_timeout(sm, eapol_ptk_1_of_4_retry);
+}
+
 static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 					const struct eapol_key *ek)
 {
@@ -820,10 +933,12 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 	uint8_t mic[16];
 	uint8_t *ies;
 	size_t ies_len;
-	const uint8_t *own_ie = sm->handshake->own_ie;
+	const uint8_t *own_ie = sm->handshake->supplicant_ie;
 	const uint8_t *pmkid;
 	struct ie_rsn_info rsn_info;
 
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
 	if (!eapol_verify_ptk_1_of_4(ek))
 		goto error_unspecified;
 
@@ -861,6 +976,8 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 
 		if (handshake_state_get_pmkid(sm->handshake, own_pmkid) &&
 				memcmp(pmkid, own_pmkid, 16)) {
+			l_debug("Authenticator sent a PMKID that didn't match");
+
 			/*
 			 * If the AP has a different PMKSA from ours and we
 			 * have means to create a new PMKSA through EAP then
@@ -869,7 +986,18 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 			if (sm->eap) {
 				send_eapol_start(NULL, sm);
 				return;
-			} else
+			}
+
+			/*
+			 * Some APs are known to send a PMKID KDE with all
+			 * zeros for the PMKID.  Likely we can still
+			 * successfully negotiate a handshake, so ignore this
+			 * for now and treat it as if the PMKID KDE was not
+			 * included
+			 */
+			if (util_mem_is_zero(pmkid, 16))
+				l_debug("PMKID is all zero, ignoring");
+			else
 				goto error_unspecified;
 		}
 	}
@@ -927,7 +1055,8 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 
 	ptk = handshake_state_get_ptk(sm->handshake);
 
-	if (!eapol_calculate_mic(ptk->kck, step2, mic)) {
+	if (!eapol_calculate_mic(sm->handshake->akm_suite, ptk->kck,
+			step2, mic)) {
 		l_info("MIC calculation failed. "
 			"Ensure Kernel Crypto is available.");
 		l_free(step2);
@@ -949,7 +1078,108 @@ error_unspecified:
 	handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
 }
 
-const uint8_t *eapol_find_rsne(const uint8_t *data, size_t data_len,
+#define EAPOL_PAIRWISE_UPDATE_COUNT 3
+
+/* 802.11-2016 Section 12.7.6.4 */
+static void eapol_send_ptk_3_of_4(struct eapol_sm *sm)
+{
+	uint32_t ifindex = sm->handshake->ifindex;
+	uint8_t frame_buf[512];
+	uint8_t key_data_buf[128];
+	struct eapol_key *ek = (struct eapol_key *) frame_buf;
+	size_t key_data_len;
+	enum crypto_cipher cipher = ie_rsn_cipher_suite_to_cipher(
+				sm->handshake->pairwise_cipher);
+	enum crypto_cipher group_cipher = ie_rsn_cipher_suite_to_cipher(
+				sm->handshake->group_cipher);
+	const struct crypto_ptk *ptk = (struct crypto_ptk *) sm->handshake->ptk;
+	struct ie_rsn_info rsn;
+	uint8_t *rsne;
+
+	sm->replay_counter++;
+
+	memset(ek, 0, sizeof(struct eapol_key));
+	ek->header.protocol_version = sm->protocol_version;
+	ek->header.packet_type = 0x3;
+	ek->descriptor_type = EAPOL_DESCRIPTOR_TYPE_80211;
+	/* Must be HMAC-SHA1-128 + AES when using CCMP with PSK or 8021X */
+	ek->key_descriptor_version = EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES;
+	ek->key_type = true;
+	ek->install = true;
+	ek->key_ack = true;
+	ek->key_mic = true;
+	ek->secure = true;
+	ek->encrypted_key_data = true;
+	ek->key_length = L_CPU_TO_BE16(crypto_cipher_key_len(cipher));
+	ek->key_replay_counter = L_CPU_TO_BE64(sm->replay_counter);
+	memcpy(ek->key_nonce, sm->handshake->anonce, sizeof(ek->key_nonce));
+	memcpy(ek->key_rsc, sm->handshake->gtk_rsc, 6);
+	ek->key_rsc[6] = 0;
+	ek->key_rsc[7] = 0;
+
+	/*
+	 * Just one RSNE in Key Data as we only set one cipher in ap->ciphers
+	 * currently.
+	 */
+
+	memset(&rsn, 0, sizeof(rsn));
+	rsn.akm_suites = IE_RSN_AKM_SUITE_PSK;
+	rsn.pairwise_ciphers = sm->handshake->pairwise_cipher;
+	rsn.group_cipher = sm->handshake->group_cipher;
+
+	rsne = key_data_buf;
+	if (!ie_build_rsne(&rsn, rsne))
+		return;
+
+	key_data_len = rsne[1] + 2;
+
+	if (group_cipher) {
+		uint8_t *gtk_kde = key_data_buf + key_data_len;
+
+		handshake_util_build_gtk_kde(group_cipher,
+						sm->handshake->gtk,
+						sm->handshake->gtk_index,
+						gtk_kde);
+		key_data_len += gtk_kde[1] + 2;
+	}
+
+	if (!eapol_encrypt_key_data(ptk->kek, key_data_buf,
+					key_data_len, ek))
+		return;
+
+	key_data_len = L_BE16_TO_CPU(ek->key_data_len);
+	ek->header.packet_len = L_CPU_TO_BE16(sizeof(struct eapol_key) +
+						key_data_len - 4);
+
+	if (!eapol_calculate_mic(sm->handshake->akm_suite, ptk->kck, ek,
+			ek->key_mic_data))
+		return;
+
+	l_debug("STA: "MAC" retries=%u", MAC_STR(sm->handshake->spa),
+			sm->frame_retry);
+
+	__eapol_tx_packet(ifindex, sm->handshake->spa, ETH_P_PAE,
+				(struct eapol_frame *) ek, false);
+}
+
+static void eapol_ptk_3_of_4_retry(struct l_timeout *timeout,
+						void *user_data)
+{
+	struct eapol_sm *sm = user_data;
+
+	if (sm->frame_retry >= EAPOL_PAIRWISE_UPDATE_COUNT) {
+		handshake_failed(sm, MMPDU_REASON_CODE_4WAY_HANDSHAKE_TIMEOUT);
+		return;
+	}
+
+	eapol_send_ptk_3_of_4(sm);
+
+	eapol_set_key_timeout(sm, eapol_ptk_3_of_4_retry);
+
+	l_debug("attempt %i", sm->frame_retry);
+}
+
+static const uint8_t *eapol_find_rsne(const uint8_t *data, size_t data_len,
 				const uint8_t **optional)
 {
 	struct ie_tlv_iter iter;
@@ -975,6 +1205,61 @@ const uint8_t *eapol_find_rsne(const uint8_t *data, size_t data_len,
 	return first;
 }
 
+/* 802.11-2016 Section 12.7.6.3 */
+static void eapol_handle_ptk_2_of_4(struct eapol_sm *sm,
+					const struct eapol_key *ek)
+{
+	const uint8_t *rsne;
+	enum crypto_cipher cipher;
+	size_t ptk_size;
+	uint8_t ptk_buf[64];
+	struct crypto_ptk *ptk = (struct crypto_ptk *) ptk_buf;
+	const uint8_t *aa = sm->handshake->aa;
+
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
+	if (!eapol_verify_ptk_2_of_4(ek))
+		return;
+
+	if (L_BE64_TO_CPU(ek->key_replay_counter) != sm->replay_counter)
+		return;
+
+	cipher = ie_rsn_cipher_suite_to_cipher(sm->handshake->pairwise_cipher);
+	ptk_size = sizeof(struct crypto_ptk) + crypto_cipher_key_len(cipher);
+
+	if (!crypto_derive_pairwise_ptk(sm->handshake->pmk, sm->handshake->spa,
+					aa, sm->handshake->anonce,
+					ek->key_nonce, ptk, ptk_size, false))
+		return;
+
+	if (!eapol_verify_mic(sm->handshake->akm_suite, ptk->kck, ek))
+		return;
+
+	/*
+	 * 12.7.6.3 b) 2) "the Authenticator checks that the RSNE bitwise
+	 * matches that from the (Re)Association Request frame.
+	 */
+	rsne = eapol_find_rsne(ek->key_data,
+				L_BE16_TO_CPU(ek->key_data_len), NULL);
+	if (!rsne || rsne[1] != sm->handshake->supplicant_ie[1] ||
+			memcmp(rsne + 2, sm->handshake->supplicant_ie + 2,
+				rsne[1])) {
+
+		handshake_failed(sm, MMPDU_REASON_CODE_IE_DIFFERENT);
+		return;
+	}
+
+	memcpy(sm->handshake->ptk, ptk_buf, ptk_size);
+	memcpy(sm->handshake->snonce, ek->key_nonce,
+			sizeof(sm->handshake->snonce));
+	sm->handshake->have_snonce = true;
+	sm->handshake->ptk_complete = true;
+
+	sm->frame_retry = 0;
+
+	eapol_ptk_3_of_4_retry(NULL, sm);
+}
+
 static const uint8_t *eapol_find_wpa_ie(const uint8_t *data, size_t data_len)
 {
 	struct ie_tlv_iter iter;
@@ -1010,6 +1295,8 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 	uint8_t gtk_key_index;
 	uint8_t igtk_key_index;
 
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
 	if (!eapol_verify_ptk_3_of_4(ek, sm->handshake->wpa_ie)) {
 		handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
 		return;
@@ -1042,7 +1329,7 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 	if (!rsne)
 		goto error_ie_different;
 
-	if (!handshake_util_ap_ie_matches(rsne, sm->handshake->ap_ie,
+	if (!handshake_util_ap_ie_matches(rsne, sm->handshake->authenticator_ie,
 						sm->handshake->wpa_ie))
 		goto error_ie_different;
 
@@ -1198,7 +1485,8 @@ retransmit:
 
 	ptk = handshake_state_get_ptk(sm->handshake);
 
-	if (!eapol_calculate_mic(ptk->kck, step4, mic)) {
+	if (!eapol_calculate_mic(sm->handshake->akm_suite, ptk->kck,
+			step4, mic)) {
 		l_free(step4);
 		handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
 		return;
@@ -1211,7 +1499,14 @@ retransmit:
 	if (sm->handshake->ptk_complete)
 		return;
 
-	handshake_state_install_ptk(sm->handshake);
+	/*
+	 * For WPA1 the group handshake should be happening after we set the
+	 * ptk, this flag tells netdev to wait for the gtk/igtk before
+	 * completing the connection.
+	 */
+	if (!gtk && sm->handshake->group_cipher !=
+			IE_RSN_CIPHER_SUITE_NO_GROUP_TRAFFIC)
+		sm->handshake->wait_for_gtk = true;
 
 	if (gtk)
 		eapol_install_gtk(sm, gtk_key_index, gtk, gtk_len, ek->key_rsc);
@@ -1219,6 +1514,8 @@ retransmit:
 	if (igtk)
 		eapol_install_igtk(sm, igtk_key_index, igtk, igtk_len);
 
+	handshake_state_install_ptk(sm->handshake);
+
 	if (rekey_offload)
 		rekey_offload(sm->handshake->ifindex, ptk->kek, ptk->kck,
 				sm->replay_counter, sm->user_data);
@@ -1232,6 +1529,29 @@ error_ie_different:
 	handshake_failed(sm, MMPDU_REASON_CODE_IE_DIFFERENT);
 }
 
+/* 802.11-2016 Section 12.7.6.5 */
+static void eapol_handle_ptk_4_of_4(struct eapol_sm *sm,
+					const struct eapol_key *ek)
+{
+	const struct crypto_ptk *ptk = (struct crypto_ptk *) sm->handshake->ptk;
+
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
+	if (!eapol_verify_ptk_4_of_4(ek, false))
+		return;
+
+	if (L_BE64_TO_CPU(ek->key_replay_counter) != sm->replay_counter)
+		return;
+
+	if (!eapol_verify_mic(sm->handshake->akm_suite, ptk->kck, ek))
+		return;
+
+	l_timeout_remove(sm->timeout);
+	sm->timeout = NULL;
+
+	handshake_state_install_ptk(sm->handshake);
+}
+
 static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 					const struct eapol_key *ek,
 					const uint8_t *decrypted_key_data,
@@ -1311,7 +1631,8 @@ static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 
 	ptk = handshake_state_get_ptk(sm->handshake);
 
-	if (!eapol_calculate_mic(ptk->kck, step2, mic)) {
+	if (!eapol_calculate_mic(sm->handshake->akm_suite, ptk->kck,
+			step2, mic)) {
 		l_free(step2);
 		handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
 		return;
@@ -1411,7 +1732,7 @@ static void eapol_key_handle(struct eapol_sm *sm,
 		if (!sm->handshake->have_snonce)
 			return;
 
-		if (!eapol_verify_mic(ptk->kck, ek))
+		if (!eapol_verify_mic(sm->handshake->akm_suite, ptk->kck, ek))
 			return;
 	}
 
@@ -1421,8 +1742,9 @@ static void eapol_key_handle(struct eapol_sm *sm,
 		if (!sm->handshake->have_snonce)
 			return;
 
-		decrypted_key_data = eapol_decrypt_key_data(ptk->kek, ek,
-						&key_data_len);
+		decrypted_key_data = eapol_decrypt_key_data(
+					sm->handshake->akm_suite, ptk->kek,
+					ek, &key_data_len);
 		if (!decrypted_key_data)
 			return;
 	} else
@@ -1575,6 +1897,54 @@ void eapol_sm_set_require_handshake(struct eapol_sm *sm, bool enabled)
 		sm->use_eapol_start = false;
 }
 
+static void eapol_auth_key_handle(struct eapol_sm *sm,
+				const struct eapol_frame *frame)
+{
+	size_t frame_len = 4 + L_BE16_TO_CPU(frame->header.packet_len);
+	const struct eapol_key *ek = eapol_key_validate((const void *) frame,
+							frame_len);
+
+	if (!ek)
+		return;
+
+	/* Wrong direction */
+	if (ek->key_ack)
+		return;
+
+	if (ek->request)
+		return; /* Not supported */
+
+	if (!sm->handshake->have_anonce)
+		return; /* Not expecting an EAPoL-Key yet */
+
+	if (!sm->handshake->ptk_complete)
+		eapol_handle_ptk_2_of_4(sm, ek);
+	else
+		eapol_handle_ptk_4_of_4(sm, ek);
+}
+
+static void eapol_rx_auth_packet(uint16_t proto, const uint8_t *from,
+				const struct eapol_frame *frame,
+				void *user_data)
+{
+	struct eapol_sm *sm = user_data;
+
+	if (!sm->protocol_version)
+		sm->protocol_version = frame->header.protocol_version;
+
+	switch (frame->header.packet_type) {
+	case 3: /* EAPOL-Key */
+		eapol_auth_key_handle(sm, frame);
+		break;
+
+	default:
+		l_error("Authenticator received unknown packet type %i from %s",
+			frame->header.packet_type,
+			util_address_to_string(from));
+		return;
+	}
+}
+
 static void eapol_rx_packet(uint16_t proto, const uint8_t *from,
 				const struct eapol_frame *frame,
 				void *user_data)
@@ -1678,11 +2048,6 @@ void __eapol_set_tx_user_data(void *user_data)
 	tx_user_data = user_data;
 }
 
-void __eapol_set_deauthenticate_func(eapol_deauthenticate_func_t func)
-{
-	deauthenticate = func;
-}
-
 void __eapol_set_rekey_offload_func(eapol_rekey_offload_func_t func)
 {
 	rekey_offload = func;
@@ -1692,7 +2057,16 @@ void eapol_register(struct eapol_sm *sm)
 {
 	l_queue_push_head(state_machines, sm);
 
-	sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
+	if (sm->handshake->authenticator) {
+		sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
+						eapol_rx_auth_packet, sm);
+
+		sm->started = true;
+
+		/* kick off handshake */
+		eapol_ptk_1_of_4_retry(NULL, sm);
+	} else
+		sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
 						eapol_rx_packet, sm);
 }
 
@@ -1748,41 +2122,6 @@ eap_error:
 	return false;
 }
 
-struct eapol_frame_watch {
-	uint32_t ifindex;
-	struct watchlist_item super;
-};
-
-static void eapol_frame_watch_free(struct watchlist_item *item)
-{
-	struct eapol_frame_watch *efw =
-		container_of(item, struct eapol_frame_watch, super);
-
-	l_free(efw);
-}
-
-static const struct watchlist_ops eapol_frame_watch_ops = {
-	.item_free = eapol_frame_watch_free,
-};
-
-uint32_t eapol_frame_watch_add(uint32_t ifindex,
-				eapol_frame_watch_func_t handler,
-				void *user_data)
-{
-	struct eapol_frame_watch *efw;
-
-	efw = l_new(struct eapol_frame_watch, 1);
-	efw->ifindex = ifindex;
-
-	return watchlist_link(&frame_watches, &efw->super,
-				handler, user_data, NULL);
-}
-
-bool eapol_frame_watch_remove(uint32_t id)
-{
-	return watchlist_remove(&frame_watches, id);
-}
-
 struct preauth_sm {
 	uint32_t ifindex;
 	uint8_t aa[6];
