@@ -41,7 +41,6 @@ typedef void (*scan_freq_set_func_t)(uint32_t freq, void *userdata);
 
 struct scan_freq_set;
 struct ie_rsn_info;
-enum ie_bss_capability;
 
 struct scan_bss {
 	uint8_t addr[6];
@@ -73,6 +72,18 @@ struct scan_parameters {
 	const char *ssid;	/* Used for direct probe request */
 };
 
+static inline int scan_bss_addr_cmp(const struct scan_bss *a1,
+					const struct scan_bss *a2)
+{
+        return memcmp(a1->addr, a2->addr, sizeof(a1->addr));
+}
+
+static inline bool scan_bss_addr_eq(const struct scan_bss *a1,
+					const struct scan_bss *a2)
+{
+        return !memcmp(a1->addr, a2->addr, sizeof(a1->addr));
+}
+
 uint32_t scan_passive(uint32_t ifindex, scan_trigger_func_t trigger,
 			scan_notify_func_t notify, void *userdata,
 			scan_destroy_func_t destroy);
@@ -97,7 +108,7 @@ void scan_sched_start(struct l_genl_family *nl80211, uint32_t ifindex,
 void scan_bss_free(struct scan_bss *bss);
 int scan_bss_rank_compare(const void *a, const void *b, void *user);
 
-int scan_bss_get_rsn_info(struct scan_bss *bss, struct ie_rsn_info *info);
+int scan_bss_get_rsn_info(const struct scan_bss *bss, struct ie_rsn_info *info);
 
 uint8_t scan_freq_to_channel(uint32_t freq, enum scan_band *out_band);
 uint32_t scan_channel_to_freq(uint8_t channel, enum scan_band band);
